@@ -24,9 +24,6 @@
 #ifndef __DETECT_FLAGS_H__
 #define __DETECT_FLAGS_H__
 
-#include "decode-events.h"
-#include "decode-ipv4.h"
-#include "decode-tcp.h"
 
 /**
  * \struct DetectFlagsData_
@@ -41,7 +38,7 @@
 typedef struct DetectFlagsData_ {
     uint8_t flags;  /**< TCP flags */
     uint8_t modifier; /**< !(1) +(2) *(3) modifiers */
-    uint8_t ignored_flags;  /**< Ignored TCP flags defined by modifer , */
+    uint8_t ignored_flags; /**< Ignored TCP flags defined by modifier , */
 } DetectFlagsData;
 
 /**
